#!/usr/bin/env python
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Author: "Chris Ward" <cward@redhat.com>

# ATTENTION: this is the main interface for clients!
from metrique.core_api import pyclient
pyclient  # ... avoid 'pep8' 'import but not used' error
